/***************************************************************************
                           Mutex.h  -  description
                             -------------------
   Provides a portable mutex API.
                             -------------------
    begin                : Sat Jul 17 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __MUTEX_H__
#define __MUTEX_H__

#if defined(POSIX)
/************************Start POSIX Version************************/
# include <stdlib.h>
# include <pthread.h>

/* Causes the current thread to lock the mutex (m) if
   available.  Thread will block only if wait flag is set.
   Returns non-zero value if thread gets lock to (m). */
#define lock_MUTEX(m, wait) \
   ( (wait) ? \
     (!pthread_mutex_lock((pthread_mutex_t *)m)) : \
     (!pthread_mutex_trylock((pthread_mutex_t *)m)) )
/* Causes the current thread to release the lock to (m).
   Returns non-zero value if thread owns lock to (m). */
#define unlock_MUTEX(m) \
   (!pthread_mutex_unlock((pthread_mutex_t *)m))
/* Destroys the mutex (m). */
#define del_MUTEX(m) \
   if (m) { \
      pthread_mutex_destroy((pthread_mutex_t *)m); \
      free(m); \
   }

/*************************End POSIX Version*************************/
#elif defined(WIN32)
/************************Start WIN32 Version************************/
# ifndef WIN32_LEAN_AND_MEAN
#  define WIN32_LEAN_AND_MEAN
# endif
# include <windows.h>

/* Causes the current thread to lock the mutex (m) if
   available.  Thread will block only if wait flag is set.
   Returns non-zero value if thread gets lock to (m). */
#define lock_MUTEX(m, wait) \
   (WaitForSingleObject((HANDLE)m, ((wait)?INFINITE:0)) == WAIT_OBJECT_0)
/* Causes the current thread to release the lock to (m).
   Returns non-zero value if thread owns lock to (m). */
#define unlock_MUTEX(m) \
   (ReleaseMutex((HANDLE)m) == TRUE)
/* Destroys the mutex (m). */
#define del_MUTEX(m) \
   if (m) CloseHandle((HANDLE)m)

/*************************End WIN32 Version*************************/
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef void * MUTEX;

/* Public Functions */
/* Creates a new mutex.  The current thread can get the
   lock for the mutex immediately by setting (lockMUTEX). */
MUTEX new_MUTEX(int lockMUTEX);

/* The rest of these functions are implemented using the macros above. */
#ifdef IGNORE_THESE
/* Causes the current thread to lock the mutex (m) if
   available.  Thread will block only if wait flag is set.
   Returns non-zero value if thread gets lock to (m). */
int   lock_MUTEX(MUTEX m, int wait);
/* Causes the current thread to release the lock to (m).
   Returns non-zero value if thread owns lock to (m). */
int   unlock_MUTEX(MUTEX m);
/* Destroys the mutex (m). */
void  del_MUTEX(MUTEX m);
#endif

#ifdef __cplusplus
};
#endif

#endif

