/***************************************************************************
                          Timer.c  -  description
                             -------------------
   You'll only find the TIMER_diff(), TIMER_add(), and TIMER_getdbl
     functions here.                             
                             -------------------
    begin                : Sat Jul 17 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>

#include "Timer.h"

#ifdef WIN32

# include "Win32/Win32_Timer.c"

#endif

/* Calculates the difference between t2 and t1 (t1 = t2 - t1).
   Result is stored in t1.
   Returns 0 if successful. */
int TIMER_diff(TIME *t1, const TIME *t2)
{
   if ( (!t1)||(!t2) )
      return -1;

   t1->tv_sec  = t2->tv_sec - t1->tv_sec;
   t1->tv_nsec = t2->tv_nsec - t1->tv_nsec;
   if ( (t1->tv_sec > 0)&&(t1->tv_nsec < 0) ) {
      (t1->tv_sec)--;
      t1->tv_nsec += 1000000000;
   }
   else if ( (t1->tv_sec < 0)&&(t1->tv_nsec > 0) ) {
      (t1->tv_sec)++;
      t1->tv_nsec -= 1000000000;
   }
   return 0;
}

/* Calculates the sum of t1 and t2 (t1 += t2).
   Result is stored in t1.
   Returns 0 if successful. */
int TIMER_add(TIME *t1, const TIME *t2)
{
   if ( (!t1)||(!t2) )
      return -1;

   t1->tv_sec  += t2->tv_sec;
   t1->tv_nsec += t2->tv_nsec;
   if (labs(t1->tv_nsec) >= 1000000000) {
      long extra_sec = t1->tv_nsec / 1000000000;

      t1->tv_sec  += extra_sec;
      t1->tv_nsec -= extra_sec * 1000000000;
   }
   return 0;
}

/* Returns the value of t as a double value */
double TIMER_getdbl(TIME *t)
{
   if (t) {
      double dbl_t = (double)t->tv_sec;

      dbl_t += (double)t->tv_nsec / 1000000000.0; 
      return dbl_t;
   }
   return 0.0;
}
