/***************************************************************************
                          MPEG2_SYS.c  -  description
                             -------------------
    begin                : Sun Aug 8 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#include <stdio.h>*/
#include <stdlib.h>
#include <string.h>

#include "common/Timer.h"
#include "AVindex.h"
#include "MPEG2_SYS.h"
#include "commands.h"
#include "PES_PKT.h"
#include "PSI_module.h"
#include "PAT_module.h"
#include "PMT_module.h"
#include "PES_module.h"
#ifndef _NO_VIDEO_
# include "mpeg2_video/MPEG2_video_module.h"
# include "mpeg1_video/MPEG1_video_module.h"
# include "vc1_video/VC1_video_module.h"
#endif
#ifndef _NO_AUDIO_
# include "ac3_audio/AC3_audio_module.h"
# include "mpeg_audio/MPEG_audio_module.h"
# include "sound/sound.h"
#endif


/*#define ALIGN_BUFFER (   2*1024)*/

MPEG2_SYS sys = { 0, 0,
#ifdef MPEG2_SYS_CLK64
                  0,
#endif
                  /* PAT module functions */
                  PAT_PSI_MODULE_add_entry,
                  PAT_PSI_MODULE_get_pid,
                  PAT_PSI_MODULE_get_program,

                  /* PMT module functions */
                  PMT_PSI_MODULE_PMT_changed,
                  PMT_PSI_MODULE_add_entry,
                  PMT_PSI_MODULE_get_PMT,
                          
                  {  /* MPEG2_PSI_PES_MODULEs */
#ifndef _NO_VIDEO_
                     MPEG2_VIDEO_MODULE,
                     MPEG1_VIDEO_MODULE,
                     VC1_VIDEO_MODULE,
#endif
#ifndef _NO_AUDIO_
                     AC3_AUDIO_MODULE,
                     MPEG_AUDIO_MODULE,
#endif
                     PMT_PSI_MODULE,
                     PAT_PSI_MODULE
                  },

                  {  /* MPEG2_SYS_MODULEs */
#ifndef _NO_VIDEO_
                     PES_SYS_MODULE,
                     PES_SYS_MODULE,
                     PES_SYS_MODULE,
#endif
#ifndef _NO_AUDIO_
                     PES_SYS_MODULE,
                     PES_SYS_MODULE,
#endif
                     PSI_SYS_MODULE,
                     PSI_SYS_MODULE
                  },

                  {  /* ACTIVE PIDs */
#ifndef _NO_VIDEO_
                     TS_PKT_NUL_ID,
                     TS_PKT_NUL_ID,
                     TS_PKT_NUL_ID,
#endif
#ifndef _NO_AUDIO_
                     TS_PKT_NUL_ID,
                     TS_PKT_NUL_ID,
#endif
                     TS_PKT_NUL_ID,
                     TS_PKT_PAT_ID
                  },

                  /* index value of the PAT module */
                  (VIDEO_DECODERS+AUDIO_DECODERS+1),
                  
                  /* index value of the PMT module */
                  (VIDEO_DECODERS+AUDIO_DECODERS),
                  
                  0,        /* index value of a PES module */
                  0,        /* program number to play */
                  
                  NULL,
                  NULL,
                  {CMD_PLAY, 0, 0, 0, 0, 0},
#ifdef _RECORD_STATS_
                  0, 0, 0.0,
#endif
                  NULL, NULL, NULL
                };

                            
/* Returns the offset needed to sync the system clock with the program clock.
      INPUT: PCR - Program clock reference value.
      RETURN: offset value. */
CLK_VAL MPEG2_get_offset(CLK_VAL PCR)
{
   TIME    curtime;
   CLK_VAL sys_clk;

   TIMER_curtime(&curtime);
   sys_clk  = (CLK_VAL)curtime.tv_nsec / (1000000000 / MPEG2_SYS_CLK_FREQ);
   sys_clk += (CLK_VAL)curtime.tv_sec  * MPEG2_SYS_CLK_FREQ;
#if (MPEG2_SYS_SLOW_DOWN > 1)
   return (PCR - sys_clk) / MPEG2_SYS_SLOW_DOWN;
#else
   return (PCR - sys_clk);
#endif
}

/* Returns the value of the program clock.
      INPUT: offset - Value needed to sync the system clock with the
                        program clock.
      RETURN: Current program clock value. */
CLK_VAL MPEG2_prog_clk(CLK_VAL offset)
{
   TIME    curtime;
   CLK_VAL sys_clk;

   TIMER_curtime(&curtime);
   sys_clk  = (CLK_VAL)curtime.tv_nsec / (1000000000 / MPEG2_SYS_CLK_FREQ);
   sys_clk += (CLK_VAL)curtime.tv_sec  * MPEG2_SYS_CLK_FREQ;
#if (MPEG2_SYS_SLOW_DOWN > 1)
   sys_clk /= MPEG2_SYS_SLOW_DOWN;
#endif
   sys_clk += offset;
   return sys_clk;
}

/* Allocates, and initializes memory for all modules.
      RETURN: Success   - Non zero value.
              No memory - Zero. */
int MPEG2_SYS_alloc()
{
   int    i;
   size_t total_mem, mem_needed;
   u8    *ptr;
   
   /* initialize syncronization objects */
   sys.m  = new_MUTEX(0);
   sys.cv = new_CONDVAR();
   if ( (!sys.m)||(!sys.cv) ) {
      i = MPEG2_SYS_PIDS+1;
      total_mem = 0;
   }
   else {
      i = 0;
#if defined(ALIGN_BUFFER)
      total_mem = (BUFFER_SIZE*NUM_BUFFERS) + (ALIGN_BUFFER-1);
#elif defined(ALIGN_BYTES)
      total_mem = (BUFFER_SIZE*NUM_BUFFERS) + (ALIGN_BYTES-1);
#else
      total_mem = (BUFFER_SIZE*NUM_BUFFERS);
#endif
   }

   /* find out how much memory is needed */
   for (; i < MPEG2_SYS_PIDS; i++) {
      if (sys.psi_pes[i].common.mem_needed) {
         mem_needed = sys.psi_pes[i].common.mem_needed();
         total_mem += mem_needed;
#ifdef ALIGN_BYTES
         if (mem_needed & (ALIGN_BYTES-1))
            total_mem += ALIGN_BYTES - (mem_needed & (ALIGN_BYTES-1));
#endif
      }
      if (sys.module[i].common.mem_needed) {
         mem_needed = sys.module[i].common.mem_needed();
         total_mem += mem_needed;
#ifdef ALIGN_BYTES
         if (mem_needed & (ALIGN_BYTES-1))
            total_mem += ALIGN_BYTES - (mem_needed & (ALIGN_BYTES-1));
#endif
      }
   }

#ifndef _SPEED_TEST_
   total_mem += AVindex_mem_needed();
#endif

   /* allocate memory */   
   if ( (i != MPEG2_SYS_PIDS)||(!(sys.mem = malloc(total_mem))) ) {
      MPEG2_SYS_free();
      return 0;
   }
   
   /* initialize buffer pointers */
   sys.buf = sys.mem;
#if defined(ALIGN_BUFFER)
   if ((u32)sys.mem & (ALIGN_BUFFER-1))
      sys.buf += ALIGN_BUFFER - ((u32)sys.mem & (ALIGN_BUFFER-1));
#elif defined(ALIGN_BYTES)
   if ((u32)sys.mem & (ALIGN_BYTES-1))
      sys.buf += ALIGN_BYTES - ((u32)sys.mem & (ALIGN_BYTES-1));
#endif
/*   printf("mem=0x%08X  buf=0x%08X  total_mem=%ubytes\n", 
          (u32)sys.mem, (u32)sys.buf, total_mem);*/
   sys.last_buf = sys.buf + (BUFFER_SIZE * (NUM_BUFFERS-1));

   /* initialize modules */
   ptr = sys.buf + (BUFFER_SIZE * NUM_BUFFERS);
   for (i = 0; i < MPEG2_SYS_PIDS; i++) {
      if ( (sys.psi_pes[i].common.mem_needed)&&
           ((mem_needed = sys.psi_pes[i].common.mem_needed())) ) {
#if defined(ALIGN_BYTES)
/*         printf("ptr=0x%08X  ", (u32)ptr);*/
         if ((u32)ptr & (ALIGN_BYTES-1))
            ptr += ALIGN_BYTES - ((u32)ptr & (ALIGN_BYTES-1));
/*         printf("psi_pes[%d]=0x%08X\n", i, (u32)ptr);*/
#endif
         sys.psi_pes[i].common.mem = ptr;
         memset(ptr, 0, mem_needed);
         ptr += mem_needed;
      }

      if ( (sys.module[i].common.mem_needed)&&
           ((mem_needed = sys.module[i].common.mem_needed())) ) {
#if defined(ALIGN_BYTES)
/*         printf("ptr=0x%08X  ", (u32)ptr);*/
         if ((u32)ptr & (ALIGN_BYTES-1))
            ptr += ALIGN_BYTES - ((u32)ptr & (ALIGN_BYTES-1));
/*         printf(" module[%d]=0x%08X\n", i, (u32)ptr);*/
#endif
         sys.module[i].common.mem = ptr;
         memset(ptr, 0, mem_needed);
         ptr += mem_needed;
      }
      
      if (sys.module[i].psi_pes_module) {
         sys.module[i].psi_pes_module(sys.module[i].common.mem,
                                      (sys.psi_pes + i));
      }
      if (sys.module[i].common.init) {
         sys.module[i].common.init(sys.module[i].common.mem,
                                   &(sys.data));
      }
   }
   
#ifndef _SPEED_TEST_
# if defined(ALIGN_BYTES)
   if ((u32)ptr & (ALIGN_BYTES-1))
      ptr += ALIGN_BYTES - ((u32)ptr & (ALIGN_BYTES-1));
# endif
   AVindex_set_mem(ptr);
#endif
   return 1;
}

/* Cleans up, and frees the memory allocated in MPEG2_SYS_alloc(). */
void MPEG2_SYS_free()
{
   if (sys.mem) {
      int i;

      /* finalize modules */
      for (i = 0; i < MPEG2_SYS_PIDS; i++) {
         if (sys.module[i].common.done)
            sys.module[i].common.done(sys.module[i].common.mem);
         sys.psi_pes[i].common.mem =
         sys.module[i].common.mem  = NULL;
      }

      /* free memory */      
      free(sys.mem);
      sys.mem      =
      sys.buf      =
      sys.last_buf = NULL;
#ifndef _SPEED_TEST_
      AVindex_set_mem(NULL);
#endif
   }

   /* free syncronization objects */
   if (sys.cv) {
      del_CONDVAR(sys.cv);
      sys.cv = NULL;
   }
   if (sys.m) {
      del_MUTEX(sys.m);
      sys.m = NULL;
   }
}

/* Uses the player function to play the content in the given file.
      INPUT: fp     - Points to the file that will be played.
             player - Function that can decode the content in
                        given file.
      RETURN: Success - A non-zero value
              Failure - 0 */
int MPEG2_SYS_Play(FILE_PTR fp, THREAD_FUNC player)
{
#ifndef _SPEED_TEST_
   FILE_POS        f_pos, next_index_pt;
#endif
   FILE_BYTE_COUNT bytes_read;
   u8             *buf;
   u32             last_cmd;
#ifndef _NO_VIDEO_
   THREAD          v;
#endif
   THREAD          a;
   MPEG2_SYS_ARG   v_arg = {NULL, 0, 0,              VIDEO_DECODERS, 0},
                   a_arg = {NULL, 0, VIDEO_DECODERS, MPEG2_SYS_PIDS, 0};
#ifdef _RECORD_STATS_
   TIME            t1, t2;
   
   sys.stream_bytes = sys.pkt_cnt = 0;
   sys.decode_time  = 0.0;
#endif
   if (sys.data.command == CMD_STOP)
      return 1;

   sys.data.a_pos  = sys.data.v_pos  =
   sys.data.a_skip = sys.data.v_skip = 0;
/*   printf("\nA:%8d   V:%8d / %d", sys.data.a_pos, sys.data.v_pos, MPEG2_SYS_CLK_FREQ);
   fflush(stdout);*/

/*   printf("MAIN>  start\n");*/
/*   last_cmd = 0;*/
   v_arg.buf = a_arg.buf = buf = sys.buf;
#ifndef _NO_VIDEO_
   v = new_THREAD(player, &v_arg);
#endif
   a = new_THREAD(player, &a_arg);

#ifndef _NO_VIDEO_
   if ( (v)&&(a) )
#else
   if (a)
#endif
   {
      MPEG2_SYS_modules_init();
      /* send initial offset value to all modules */
      sys.offset = MPEG2_get_offset(0);
      MPEG2_SYS_decoder_pause();
#ifndef _SPEED_TEST_
      FILE_GETPOS(fp, f_pos);
      next_index_pt = AVINDEX_INTERVAL;
      AVindex_init(f_pos, 0, 0, 0, 0);
      if (sys.data.command == CMD_JUMP) {
# ifndef _NO_VIDEO_
         f_pos = MPEG2_SYS_decoder_jump(fp, &v_arg, &a_arg);
# else
         f_pos = MPEG2_SYS_decoder_jump(fp, NULL, NULL);
# endif
      }
#endif
/*
      if (sys.data.command != CMD_STOP)
         printf("MAIN>  working\n");
*/
#ifdef _RECORD_STATS_
      TIMER_curtime(&t1);
#endif
      while (sys.data.command != CMD_STOP) {
         FILE_READ(fp, buf, BUFFER_SIZE, bytes_read);
         if (bytes_read <= 1)
            break;

         lock_MUTEX(sys.m, 1);
#ifndef _NO_VIDEO_
         v_arg.bytes_read += (u32)bytes_read;
#endif
         a_arg.bytes_read += (u32)bytes_read;
         broadcast_CONDVAR(sys.cv);
         unlock_MUTEX(sys.m);

#ifndef _SPEED_TEST_
         f_pos += (FILE_POS)bytes_read;
#endif
#ifdef _RECORD_STATS_
         sys.stream_bytes += (u32)bytes_read;
#endif

         lock_MUTEX(sys.m, 1);
#ifndef _NO_VIDEO_
         while ( ( (sys.data.command != CMD_SLOW_FWD)&&
                   ( ( (v_arg.bytes_read)&&(a_arg.bytes_read) )||
                     (v_arg.bytes_read > (BUFFER_SIZE*(NUM_BUFFERS-2))+1)||
                     (a_arg.bytes_read > (BUFFER_SIZE*(NUM_BUFFERS-2))+1) ) )||
                 ( (sys.data.command == CMD_SLOW_FWD)&&
                   ( (v_arg.bytes_read > 1)||(a_arg.bytes_read > 1) ) ) )
#else
         while (a_arg.bytes_read > 1)
#endif
         {
/*            if (last_cmd < v_arg.bytes_read)
               last_cmd = v_arg.bytes_read;
            if (last_cmd < a_arg.bytes_read)
               last_cmd = a_arg.bytes_read;*/
            wait_CONDVAR(sys.cv, sys.m);
         }
#ifndef _NO_VIDEO_
         buf = (!v_arg.bytes_read) ? v_arg.buf : a_arg.buf;
#else
         buf = a_arg.buf;
#endif
#ifndef _SPEED_TEST_
         if (f_pos >= next_index_pt) {
            next_index_pt += AVINDEX_INTERVAL;
            AVindex_add(f_pos, 
                        v_arg.pos,        a_arg.pos, 
                        v_arg.bytes_read, a_arg.bytes_read);
         }
#endif
         unlock_MUTEX(sys.m);

MPEG2_SYS_Play_PAUSE_DONE:
         if (bytes_read < BUFFER_SIZE)
            break;
         else if (sys.data.command == CMD_PAUSE) {
#ifdef _RECORD_STATS_
            TIMER_curtime(&t2);
            TIMER_diff(&t1, &t2);
            sys.decode_time += TIMER_getdbl(&t1);
#endif
            MPEG2_SYS_decoder_pause(&sys);
#ifdef _RECORD_STATS_
            TIMER_curtime(&t1);
#endif
            goto MPEG2_SYS_Play_PAUSE_DONE;
         }
#ifndef _SPEED_TEST_
         else if ( (sys.data.command == CMD_JUMP)
                   &&(sys.data.a_skip < 10)
# ifndef _NO_VIDEO_
                   &&(sys.data.v_skip < 10)
# endif
                 ) {
# ifndef _NO_VIDEO_
            f_pos = MPEG2_SYS_decoder_jump(fp, &v_arg, &a_arg);
# else
            f_pos = MPEG2_SYS_decoder_jump(fp, NULL, NULL);
# endif
         }
#endif
      }
#ifdef _RECORD_STATS_
      TIMER_curtime(&t2);
      TIMER_diff(&t1, &t2);
      sys.decode_time += TIMER_getdbl(&t1);
#endif
   }

/*   printf("max buffers used = %u/%u\n", (last_cmd/BUFFER_SIZE)+2, NUM_BUFFERS);*/

   /* shutdown threads */
/*   printf("MAIN>  stop threads\n");*/
   last_cmd = sys.data.command;
   lock_MUTEX(sys.m, 1);
#ifndef _NO_VIDEO_
   while ( (v_arg.bytes_read > 1)||
           (a_arg.bytes_read > 1) )
#else
   while (a_arg.bytes_read > 1)
#endif
   {
      wait_CONDVAR(sys.cv, sys.m);  /* wait for threads to finish processing */
   }
   v_arg.bytes_read = a_arg.bytes_read = 2;
   sys.data.command = CMD_STOP;
   unlock_MUTEX(sys.m);
   
   lock_MUTEX(sys.m, 1);
#ifndef _NO_VIDEO_
   while ( (v_arg.bytes_read)||(a_arg.bytes_read) )
#else
   while (a_arg.bytes_read)
#endif
   {
      broadcast_CONDVAR(sys.cv);
      wait_CONDVAR(sys.cv, sys.m);  /* wait for threads to shutdown */
   }
   v_arg.bytes_read = a_arg.bytes_read = 0xFFFFFFFF;
   broadcast_CONDVAR(sys.cv);       /* acknowledge shutdown */
   unlock_MUTEX(sys.m);

#ifndef _NO_VIDEO_
   if (v) {
      join_THREAD(v, NULL);
      del_THREAD(v);
# ifdef _RECORD_STATS_
      sys.pkt_cnt = v_arg.pkt_cnt;
# endif
/*      printf("MAIN>  VIDEO done!\n");*/
   }
#endif
   if (a) {
      join_THREAD(a, NULL);
      del_THREAD(a);
#ifdef _RECORD_STATS_
      if (a_arg.pkt_cnt > sys.pkt_cnt)
         sys.pkt_cnt = a_arg.pkt_cnt;
#endif
/*      printf("MAIN>  AUDIO done!\n");*/
   }
   sys.data.command = last_cmd;

#ifdef _RECORD_STATS_
   TIMER_curtime(&t1);
#endif
   MPEG2_SYS_modules_done();
#ifdef _RECORD_STATS_
   TIMER_curtime(&t2);
   TIMER_diff(&t1, &t2);
   sys.decode_time += TIMER_getdbl(&t1);
#endif
/*   printf("\rA:%8d   V:%8d\n", sys.data.a_pos, sys.data.v_pos);*/
/*   printf("MAIN>  end\n");*/
#ifndef _NO_VIDEO_
   return ( (v)&&(a) );
#else
   return (int)a;
#endif
}

void MPEG2_SYS_modules_init()
{
   u32 i;

   for (i = 0; i < MPEG2_SYS_PIDS; i++) {
      if ( (i != sys.PAT_idx)&&(i != sys.PMT_idx)&&
           (sys.module[i].common.init) ) {
         sys.module[i].common.init(sys.module[i].common.mem, &(sys.data));
      }
   }
}

void MPEG2_SYS_modules_done()
{
   u32 i;

   for (i = 0; i < MPEG2_SYS_PIDS; i++) {
      if ( (i != sys.PAT_idx)&&(i != sys.PMT_idx)&&
           (sys.module[i].common.done) ) {
         sys.module[i].common.done(sys.module[i].common.mem);
      }
   }
}

void MPEG2_SYS_decoder_pause()
{
   u32 cmd, i;

   /* save the last clock value before pausing */
   sys.offset = MPEG2_prog_clk(sys.offset);

   /* pause */
#ifndef _NO_AUDIO_
   SOUND_stop(1);
#endif
   lock_MUTEX(sys.m, 1);
   while (sys.data.command == CMD_PAUSE) {
      wait_CONDVAR(sys.cv, sys.m);
   }
   unlock_MUTEX(sys.m);
#ifndef _NO_AUDIO_
   SOUND_stop(0);
#endif

   if (sys.data.command != CMD_STOP) {
      /* get the offset needed to restore the last clock value */
#if (MPEG2_SYS_SLOW_DOWN > 1)
      sys.offset = MPEG2_get_offset(sys.offset*MPEG2_SYS_SLOW_DOWN);
#else
      sys.offset = MPEG2_get_offset(sys.offset);
#endif
      /* alert all modules that we are returning from a pause */
      cmd              = sys.data.command;
      sys.data.command = CMD_PAUSE;
      for (i = 0; i < MPEG2_SYS_PIDS; i++) {
         if ( (i != sys.PAT_idx)&&(i != sys.PMT_idx)&&
              (sys.module[i].common.set_offset) ) {
            sys.module[i].common.set_offset(sys.module[i].common.mem,
                                            sys.offset);
         }
      }
      sys.data.command = cmd;
   }
}

FILE_POS MPEG2_SYS_decoder_jump(FILE_PTR       fp, 
                                MPEG2_SYS_ARG *v_arg, 
                                MPEG2_SYS_ARG *a_arg)
{
   FILE_POS f_pos, cur_pos;
   CLK_VAL  v_pos, a_pos, inc;
   u32      v_bck, a_bck;

   FILE_GETPOS(fp, cur_pos);
   
   lock_MUTEX(sys.m, 1);
   if ( (v_arg)&&(a_arg) ) {
      while ( (v_arg->bytes_read > 1)||(a_arg->bytes_read > 1) ) {
         wait_CONDVAR(sys.cv, sys.m);  /* wait for threads to finish processing */
      }
   }
   inc              = sys.jump_offset;
NEW_JUMP_REQUEST:
   sys.jump_offset  = 0;
   sys.data.command = CMD_PLAY;
   broadcast_CONDVAR(sys.cv);       /* acknowledge jump request */
   unlock_MUTEX(sys.m);

   if (!inc) {
      sys.data.a_skip = sys.data.v_skip = 0;
      goto RESET_FILEPOS;
   }
      
#ifndef _NO_VIDEO_
   a_pos = sys.data.a_skip = sys.data.a_pos + inc;
   v_pos = sys.data.v_skip = sys.data.v_pos + inc;

   if ( (!AVindex_get(&f_pos, &v_pos, &a_pos, &v_bck, &a_bck))||
#else
   a_pos = sys.data.a_skip = sys.data.a_pos + inc;
   v_pos = 0; v_bck = 0;

   if ( (!AVindex_get(&f_pos, NULL,   &a_pos, NULL, &a_bck))||
#endif
        ( (inc > 0)&&(f_pos < cur_pos) )||
        ( (inc < 0)&&(f_pos > cur_pos) ) ) {
      /* location is not in the index, so use default seek method */
      if (inc > 0) {
         sys.data.a_skip = sys.data.v_skip = inc;
      }
      else {
RESET_FILEPOS:
         sys.data.a_pos  = sys.data.v_pos  = 0;
         FILE_REWIND(fp);
         MPEG2_SYS_discontinuity_alert(v_arg, a_arg);
      }
/*      printf("Default seek\n");*/
   }
   else {
      /* location was in the index, so use quick seek method */
      FILE_SETPOS(fp, f_pos);
      sys.data.a_pos = a_pos; sys.data.a_skip -= a_pos;
      sys.data.v_pos = v_pos; sys.data.v_skip -= v_pos;
      MPEG2_SYS_discontinuity_alert(v_arg, a_arg);
      if ( (v_arg)&&(a_arg)&&((v_bck > 1)||(a_bck > 1)) ) {
         u8              *buf, *buf2;
         FILE_BYTE_COUNT  bytes_read = BUFFER_SIZE;

         /* allow the lagging thread to catch up to leading thread */
         if (v_bck) {
            buf = a_arg->buf;
            if (a_arg->buf == sys.last_buf)
               v_arg->buf = sys.buf;
            else
               v_arg->buf = a_arg->buf + BUFFER_SIZE;
         }
         else {
            buf = v_arg->buf;
            if (v_arg->buf == sys.last_buf)
               a_arg->buf = sys.buf;
            else
               a_arg->buf = v_arg->buf + BUFFER_SIZE;
         }

         do {
            lock_MUTEX(sys.m, 1);
            if (v_bck) {
               if (v_bck > BUFFER_SIZE) {
                  v_arg->bytes_read = BUFFER_SIZE;
                  v_bck            -= BUFFER_SIZE;
               }
               else {
                  bytes_read        = (FILE_BYTE_COUNT)v_bck;
                  v_arg->bytes_read = v_bck;
                  v_bck             = 0;
               }
               buf2 = v_arg->buf;
            }
            else {
               if (a_bck > BUFFER_SIZE) {
                  a_arg->bytes_read = BUFFER_SIZE;
                  a_bck            -= BUFFER_SIZE;
               }
               else {
                  bytes_read        = (FILE_BYTE_COUNT)a_bck;
                  a_arg->bytes_read = a_bck;
                  a_bck             = 0;
               }
               buf2 = a_arg->buf;
            }
            FILE_READ(fp, buf2, bytes_read, bytes_read);
            broadcast_CONDVAR(sys.cv);
            do {
               wait_CONDVAR(sys.cv, sys.m);
            } while (v_arg->bytes_read > 1);
            unlock_MUTEX(sys.m);
         } while ( (v_bck != a_bck)&&(bytes_read == BUFFER_SIZE) );
         v_arg->buf = a_arg->buf = buf;
      }
/*      printf("Quick seek\n");*/
   }
   FILE_GETPOS(fp, f_pos);
   
   lock_MUTEX(sys.m, 1);
   if (sys.data.command == CMD_JUMP) {
      if (sys.jump_offset) {
         inc  =  sys.jump_offset;
#if defined(_NO_VIDEO_)
         inc += sys.data.a_skip;
#elif defined(_NO_AUDIO_)
         inc += sys.data.v_skip;
#else
         inc += (sys.data.a_skip > sys.data.v_skip) ?
                 sys.data.a_skip : sys.data.v_skip;
#endif
      }
      else
         inc  = 0;
      cur_pos = f_pos;
      goto NEW_JUMP_REQUEST;
   }
#ifdef _NO_VIDEO_
   sys.data.v_pos  =
   sys.data.v_skip = 0;
#endif
   unlock_MUTEX(sys.m);
   
   return f_pos;
}

void MPEG2_SYS_discontinuity_alert(MPEG2_SYS_ARG *v_arg,
                                   MPEG2_SYS_ARG *a_arg)
{
   u32 i;

   if ( (v_arg)&&(a_arg) ) {
      u8  *v_buf, *a_buf;
      u32  v_bck,  a_bck;
      /* alert the ps or ts threads of the discontinuity */
      lock_MUTEX(sys.m, 1);
      v_buf = v_arg->buf;        a_buf = a_arg->buf;
      v_bck = v_arg->bytes_read; a_bck = a_arg->bytes_read;
      
      v_arg->buf        = a_arg->buf        = NULL;
      v_arg->bytes_read = a_arg->bytes_read = 0;
      broadcast_CONDVAR(sys.cv);
      do {
         wait_CONDVAR(sys.cv, sys.m);  /* wait for threads to acknowledge */
      } while ( (!v_arg->buf)||(!a_arg->buf) );
      
      v_arg->buf        = v_buf; a_arg->buf        = a_buf;
      v_arg->bytes_read = v_bck; a_arg->bytes_read = a_bck;
      unlock_MUTEX(sys.m);
   }

   /* alert all modules of an impending discontinuity in the data stream */
   for (i = 0; i < MPEG2_SYS_PIDS; i++) {
      if ( (i != sys.PAT_idx)&&(i != sys.PMT_idx)&&
           (sys.module[i].common.discontinuity != NULL) ) {
         sys.module[i].common.discontinuity(sys.module[i].common.mem);
      }
   }
}

void MPEG2_SYS_reset_PIDs()
{
   u32 i;

   for (i = 0; i < MPEG2_SYS_PIDS; i++) {
      if (i == sys.PAT_idx) {
         if (sys.module[i].common.done)
            sys.module[i].common.done(sys.module[i].common.mem);
      }
      else {
         if ( (i == sys.PMT_idx)&&(sys.module[i].common.done) )
            sys.module[i].common.done(sys.module[i].common.mem);
         sys.PID[i] = TS_PKT_NUL_ID;
      }
   }
}


#undef NUM_BUFFERS

