/***************************************************************************
                          mpeg_ts.c  -  description
                             -------------------
   Functions for handling mpeg transport streams.
                             -------------------
    begin                : Sun Aug 8 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string.h>

#include "common/Timer.h"
#include "MPEG2_SYS.h"
#include "PES_PKT.h"
#include "mpeg_ts.h"
#include "commands.h"
#include "detect_mpeg.h"


typedef struct {
   u16 vid_PID, aud_PID;
} CREATE_PROGRAM;
static int create_program(TS_PKT *ts, void *arg);

typedef struct {
   u16 *PID;
   u32  size, found;
} GET_PIDS;
static int get_PIDs(TS_PKT *ts, void *arg);

typedef struct {
   DETECT_MODULE_data mod_data;
   u16                PID, is_PES;
} STREAM_TYPE;
static int stream_type(TS_PKT *ts, void *arg);

typedef struct {
   MPEG2_SYS_MODULE *mod;
   u16               PID;
   u8                cont_cnt, last_cont_cnt;
} LOAD_PAT_PMT;
static int load_pat_pmt(TS_PKT *ts, void *arg);

static void handle_PAT_PKT(MPEG2_SYS_ARG *sys_arg);
static void handle_PMT_PKT(MPEG2_SYS_ARG *sys_arg);
#ifdef MPEG2_SYS_CLK64
static void update_offset(ADAPT_FIELD *af_ptr);
#endif

static u8  cont_cnt[MPEG2_SYS_PIDS],
           last_cont_cnt[MPEG2_SYS_PIDS];
static u32 stop_working = 2;


/* A player function to be used with the MPEG2_SYS_Play() function.
     This function can handle MPEG transport streams. */
void * THREAD_FUNC_TYPE MPEG_TS_Play(void *arg)
{
   TS_PKT         ts;
   u8             pkt[TS_PKT_SIZE];
   const u8      *ptr, *ptr2, *buf;
   u32            i, buf_size, pkt_size, no_data;
   CLK_VAL       *pos;
   MPEG2_SYS_ARG *sys_arg = (MPEG2_SYS_ARG *)arg;

/*   printf("%s> start\n", (sys_arg->si) ? "AUDIO" : "VIDEO");*/
   
   /* initialize some things before we start */
   for (i = sys_arg->si; i < sys_arg->fi; i++) {
      last_cont_cnt[i] = 0xFF;
   }
   no_data = 0;

   if (sys_arg->si == VIDEO_DECODERS) {
      handle_PAT_PKT(sys_arg);
      if (sys.PID[sys.PMT_idx] < TS_PKT_NUL_ID) {
         sys.PMT_changed(sys.psi_pes[sys.PMT_idx].common.mem);
         handle_PMT_PKT(sys_arg);
      }
      if (!stop_working) {
         for (i = sys_arg->si; i < sys.PMT_idx; i++) {
            if (sys.PID[i] < TS_PKT_NUL_ID)
               break;
         }
         if (i == sys.PMT_idx)
            no_data = 1;
      }
      pos = &(sys.data.a_pos);
   }
   else
      pos = &(sys.data.v_pos);

   /* check if there is anything to do */
   buf_size = 0;
   lock_MUTEX(sys.m, 1);
   sys_arg->pos = 0;
   if (0) {
MPEG_TS_Play_STOP_WORKING:
/*      printf("%s> stopping\n", (sys_arg->si) ? "AUDIO" : "VIDEO");*/
      /* shutdown the elementary streams */
      for (i = sys_arg->si; i < sys_arg->fi; i++) {
         sys.PID[i]       = TS_PKT_NUL_ID;
         last_cont_cnt[i] = 0xFF;
         if (sys.module[i].common.done != NULL)
            sys.module[i].common.done(sys.module[i].common.mem);
      }
      stop_working = 2;
      broadcast_CONDVAR(sys.cv); /* acknowledge the stop_working signal */
      buf_size     = no_data;
      sys_arg->pos = 0;
      goto MPEG_TS_Play_NOTHING_TO_DO;
   }
   while (sys.data.command != CMD_STOP) {
      for (i = sys_arg->si; i < sys_arg->fi; i++) {
         if (sys.PID[i] < TS_PKT_NUL_ID)
            goto MPEG_TS_Play_START_WORKING;
      }
MPEG_TS_Play_NOTHING_TO_DO:
      if (sys_arg->bytes_read > buf_size) {
         sys_arg->buf += sys_arg->bytes_read - buf_size;
         if (sys_arg->buf > sys.last_buf)
            sys_arg->buf = sys.buf;
      }
      else if (!sys_arg->buf) {
         sys_arg->buf++;
         broadcast_CONDVAR(sys.cv); /* acknowledge the discontinuity notice */
      }
      sys_arg->bytes_read =
      buf_size            = 1;
      wait_CONDVAR(sys.cv, sys.m);
      pos[0] = pos[2] = 0;
   }
   goto MPEG_TS_Play_SHUTDOWN;
MPEG_TS_Play_START_WORKING:
   if ( (sys_arg->bytes_read)&&(buf_size) )
      sys_arg->bytes_read -= buf_size;
   sys_arg->bytes_read += no_data;
   if (sys_arg->si != VIDEO_DECODERS) {
      stop_working = 0;
      broadcast_CONDVAR(sys.cv);
   }
   unlock_MUTEX(sys.m);
/*
   if (sys.data.command != CMD_STOP)
      printf("%s> working\n", (sys_arg->si) ? "AUDIO" : "VIDEO");
*/
   ptr2     = NULL;
   pkt_size = 0;

   while (sys.data.command != CMD_STOP) {
      lock_MUTEX(sys.m, 1);
      if ( (sys_arg->si != VIDEO_DECODERS)&&(stop_working) )
         goto MPEG_TS_Play_STOP_WORKING;
      sys_arg->pos = *pos;
      while (sys_arg->bytes_read <= no_data) {
         if (!sys_arg->buf) {
            sys_arg->buf++;
            ptr2     = NULL;
            pkt_size = 0;
            broadcast_CONDVAR(sys.cv); /* acknowledge the discontinuity notice */
         }
         wait_CONDVAR(sys.cv, sys.m);
         if (sys.data.command == CMD_STOP) {
/*            printf("%s> trying to end...\n", (sys_arg->si) ? "AUDIO" : "VIDEO");*/
            goto MPEG_TS_Play_SHUTDOWN;   /* shutdown */
         }
      }
      sys_arg->bytes_read -= no_data;
      if (sys_arg->bytes_read > BUFFER_SIZE) {
         buf_size             = BUFFER_SIZE;
         sys_arg->bytes_read -= BUFFER_SIZE;
      }
      else {
         buf_size             = sys_arg->bytes_read;
         sys_arg->bytes_read  = no_data;  /* ask MPEG2_SYS_Play() thread for more data */
      }
      ptr = sys_arg->buf;
      if (sys_arg->buf >= sys.last_buf)
         sys_arg->buf  = sys.buf;
      else
         sys_arg->buf += BUFFER_SIZE;
      broadcast_CONDVAR(sys.cv);
      unlock_MUTEX(sys.m);
      
      if ( (pkt_size)&&((pkt_size+buf_size) >= TS_PKT_SIZE)&&(ptr) ) {
         /* complete the last packet */
         memcpy((pkt+pkt_size), ptr, (size_t)(TS_PKT_SIZE-pkt_size));
         ptr2     = ptr + (TS_PKT_SIZE-pkt_size);
         pkt_size = buf_size - (TS_PKT_SIZE-pkt_size);
         ptr      = pkt;         /* process this packet now */
         buf_size = TS_PKT_SIZE;
      }

MPEG_TS_Play_PROCESS_PACKETS:
      buf = ptr;
      while (buf_size >= TS_PKT_SIZE) {
         if (*buf != TS_PKT_SYNC_BYTE) {
            buf = TS_PKT_find(buf, buf_size);
            if (!buf) {
               buf_size = 0;
               break;   /* this and all following packets
                           in the buffer are bad */
            }
            goto MPEG_TS_Play_PACKET_DONE;
         }

         buf = TS_PKT_read((buf+1), &ts);
         if (!buf) {
            buf = ptr + TS_PKT_SIZE;
            goto MPEG_TS_Play_PACKET_DONE;   /* this packet is bad, so skip it */
         }
#ifdef _RECORD_STATS_
         sys_arg->pkt_cnt++;
#endif
         if (ts.PID >= TS_PKT_NUL_ID)
            goto MPEG_TS_Play_PACKET_DONE;   /* ignore NULL packets */
#ifdef MPEG2_SYS_CLK64
         else if (ts.adapt_field_pos) {
            /* get the ADAPTATION FIELD */
            if (!ADAPT_FIELD_read(ts.adapt_field_pos, &af))
               goto MPEG_TS_Play_PACKET_DONE;/* ADAPTATION FIELD is corrupt */
               
            af_ptr = &af;
            if ( (ts.PID == PCR_PID)&&(ADAPT_FIELD_PCR_FLAG(af_ptr)) )
               update_offset(sys, af_ptr);
         }
         else
            af_ptr = NULL;
#endif

/*********** START TO DO SOMETHING WITH ts ***********/
         /* check if ts.PID is in the active PID list */
         for (i = sys_arg->si; i < sys_arg->fi; i++) {
            if (sys.PID[i] == ts.PID)
               break;
         }
         if (i == sys_arg->fi)
            goto MPEG_TS_Play_PACKET_DONE;   /* not interested in this packet */

         /* check for discontinuity in data stream */
         if (last_cont_cnt[i] != 0xFF) {
            if (ts.continuity_counter != (cont_cnt[i]&TS_PKT_CONT_CNT_MASK)) {
#ifdef MPEG2_SYS_CLK64
               if ( (ts.continuity_counter != last_cont_cnt[i])||
                    ( (af_ptr)&&(ADAPT_FIELD_DISCONTINUITY_FLAG(af_ptr)) ) ) {
#else
               if (ts.continuity_counter != last_cont_cnt[i]) {
#endif
                  /* counter discontinuity */
                  cont_cnt[i] = ts.continuity_counter;
                  sys.module[i].common.discontinuity(sys.module[i].common.mem);
               }
               else
                  goto MPEG_TS_Play_PACKET_DONE;   /* duplicate packet */
            }
         }
         else {
            /* this is the first packet with ts.PID == sys.PID[i] */
            cont_cnt[i] = ts.continuity_counter;
         }

         last_cont_cnt[i] = ts.continuity_counter;
         if (ts.payload)
           cont_cnt[i]++;

         sys.module[i].process_pkt(sys.module[i].common.mem, &ts);
         if (i == sys.PAT_idx) {
            handle_PAT_PKT(sys_arg);
            i = MPEG2_SYS_PIDS;
         }
         else if ( (i == sys.PMT_idx)&&
                   (sys.PMT_changed(sys.psi_pes[i].common.mem)) ) {
            handle_PMT_PKT(sys_arg);
            i = MPEG2_SYS_PIDS;
         }
         if (i == MPEG2_SYS_PIDS) {
            no_data = 0;
            if (!stop_working) {
               for (i = sys_arg->si; i < sys.PMT_idx; i++) {
                  if (sys.PID[i] < TS_PKT_NUL_ID)
                     break;
               }
               if (i == sys.PMT_idx)
                  no_data = 1;
            }
         }
/*********** STOP DOING SOMETHING WITH ts ************/

MPEG_TS_Play_PACKET_DONE:
         buf_size -= (u32)(buf - ptr);
         ptr       = buf;
      }

      if (ptr2) {
         /* process the rest of the packets in the buffer */
         ptr      = ptr2;
         ptr2     = NULL;
         buf_size = pkt_size;
         pkt_size = 0;
         goto MPEG_TS_Play_PROCESS_PACKETS;
      }
      else if ( (buf_size)&&(buf) ) {
         /* store first part of the incomplete packet */
         memcpy(pkt, buf, (size_t)buf_size);
         pkt_size = buf_size;
      }
   }

   lock_MUTEX(sys.m, 1);
MPEG_TS_Play_SHUTDOWN:
   if (sys_arg->si != VIDEO_DECODERS)
      stop_working = 2;
/*   i = 0;*/
   do {
/*      i++;*/
      sys_arg->bytes_read = 0;
      broadcast_CONDVAR(sys.cv);
      wait_CONDVAR(sys.cv, sys.m);
   } while (sys_arg->bytes_read != 0xFFFFFFFF); /* wait for everyone to stop */
   unlock_MUTEX(sys.m);
/*   printf("%s> end after %u attempts\n", (sys_arg->si) ? "AUDIO" : "VIDEO", i);*/
   exit_THREAD(NULL);
   return NULL;
}

/* Finds the first program in the PAT.
      INPUT: fp - Points to the FILE with the MPEG transport stream. 
      RETURN: Program number is stored in the global struct 
                (sys.program_number).
              Success - A non-zero value
              Failure - 0 */
void MPEG_TS_Find_Program(FILE_PTR fp)
{
   /* load the PAT, and get the first program */
   sys.program_number = 0;
   MPEG_TS_load_PAT(fp);
   sys.program_number = sys.get_program(sys.psi_pes[sys.PAT_idx].common.mem, 0);

   if (sys.program_number) {
      /* load the PMT, to make sure PAT entry is valid */
      MPEG_TS_load_PMT(fp);
      if (sys.get_PMT(sys.psi_pes[sys.PMT_idx].common.mem) == NULL)
         sys.program_number = 0;    /* PMT is missing! */
   }
}

/* Creates a program, using the first audio and video streams
     found in the given MPEG transport stream.
      INPUT: fp - Points to the FILE with the MPEG transport stream.
      RETURN: Program number, of newly created program, is stored
                in the global struct (sys.program_number).
              Success - A non-zero value
              Failure - 0 */
void MPEG_TS_Create_Program(FILE_PTR fp)
{
   FILE_POS       fp_pos;
   CREATE_PROGRAM arg      = {TS_PKT_NUL_ID, TS_PKT_NUL_ID};
   u8             vid_type = 0,
                  aud_type = 0;

   FILE_GETPOS(fp, fp_pos);
   MPEG_TS_decode(fp, (fp_pos+CMD_SEARCH_SIZE), create_program, &arg);
   FILE_SETPOS(fp, fp_pos);

   /* find correct stream type values */
   if (arg.vid_PID < TS_PKT_NUL_ID) {
      vid_type = MPEG_TS_stream_type(fp, arg.vid_PID);
      if (!vid_type)
         arg.vid_PID = TS_PKT_NUL_ID;
   }
   if (arg.aud_PID < TS_PKT_NUL_ID) {
      aud_type = MPEG_TS_stream_type(fp, arg.aud_PID);
      if (!aud_type)
         arg.aud_PID = TS_PKT_NUL_ID;
   }

   if ( ((arg.vid_PID < TS_PKT_NUL_ID)||(arg.aud_PID < TS_PKT_NUL_ID))&&
        (sys.add_PAT_entry)&&(sys.add_PMT_entry) ) {
      /* create the TS program */
      sys.program_number = 1;
      sys.add_PAT_entry(sys.psi_pes[sys.PAT_idx].common.mem,
                        sys.program_number, 0x1FFE);
      sys.add_PMT_entry(sys.psi_pes[sys.PMT_idx].common.mem,
                        sys.program_number, 0x1FFF,
                        arg.vid_PID, vid_type,
                        arg.aud_PID, aud_type);
   }
   else
      sys.program_number = 0;
}

/* Gets the stream ids and types from the transport stream.
      INPUT: fp   - Points to the FILE with the MPEG transport
                      stream.
             size - Number of elements in the PID and stream_type
                      arrays.
      OUTPUT: PID         - Array of stream ids found in the transport
                              stream.
              stream_type - Array of STREAM_TYPE_* values 
                              corresponding to the values in the
                              PID array.
      RETURN: The number of PIDs found. */
u32 MPEG_TS_get_PIDs(FILE_PTR fp, u16 *PID, u8 *stream_type, u32 size)
{
   u32      i;
   FILE_POS fp_pos;
   GET_PIDS arg = {PID, size, 0};

   memset(PID, 0xFF, sizeof(u16)*(size_t)size);
   memset(stream_type, 0, sizeof(u8)*(size_t)size);
   
   /* get PIDs first */
   FILE_GETPOS(fp, fp_pos);
   MPEG_TS_decode(fp, (fp_pos+CMD_SEARCH_SIZE), get_PIDs, &arg);
   FILE_SETPOS(fp, fp_pos);

   /* now get stream_types */
   size = arg.found;
   for (i = 0; i < size; i++) {
      if ( (PID[i] != TS_PKT_PAT_ID)&&
           (PID[i] != TS_PKT_CAT_ID)&&
           (PID[i] != TS_PKT_ATSC_SI_ID) ) {
         stream_type[i] = MPEG_TS_stream_type(fp, PID[i]);
      }
   }
   return size;
}

/* Tries to determine the type of data contained in the stream with
     the given PID.
      INPUT: fp  - Pointer to an open file that contains MPEG TS data.
             PID - PID of the stream to look for.
      RETURN: Success - One of the STREAM_TYPE_* values defined
                          in PMT_SECTION.h
              Failure - Zero. */
u8 MPEG_TS_stream_type(FILE_PTR fp, u16 PID)
{
   STREAM_TYPE          arg;
   FILE_POS             fp_pos;
   MPEG2_SYS_MODULE    *pes      = &(sys.module[sys.PES_idx]);
   MPEG2_PSI_PES_MODULE detector = DETECT_MODULE;

   memset(&arg, 0, sizeof(STREAM_TYPE));
   arg.PID             = PID;
   detector.common.mem = &(arg.mod_data);

   /* make the PES module use the detect module */
   pes->psi_pes_module(pes->common.mem, &detector);

   /* initialize pes module */
   pes->common.init(pes->common.mem, &(sys.data));

   /* determine stream type */
   FILE_GETPOS(fp, fp_pos);
   MPEG_TS_decode(fp, (fp_pos+CMD_SEARCH_SIZE), stream_type, &arg);
   FILE_SETPOS(fp, fp_pos);

   /* clean-up pes module */
   pes->common.done(pes->common.mem);

   /* make the PES module use the original module */
   pes->psi_pes_module(pes->common.mem, &(sys.psi_pes[sys.PES_idx]));

   return arg.mod_data.stream_type;
}

/* Loads the PAT.
      INPUT: fp - Pointer to an open file that contains MPEG TS data. */
void MPEG_TS_load_PAT(FILE_PTR fp)
{
   FILE_POS     fp_pos;
   LOAD_PAT_PMT arg = {&(sys.module[sys.PAT_idx]),
                       TS_PKT_PAT_ID,
                       0x00, 0xFF};

   FILE_GETPOS(fp, fp_pos);
   MPEG_TS_decode(fp, (fp_pos+CMD_SEARCH_SIZE), load_pat_pmt, &arg);
   FILE_SETPOS(fp, fp_pos);
}

/* Loads the PMT.
      INPUT: fp - Pointer to an open file that contains MPEG TS data. */
void MPEG_TS_load_PMT(FILE_PTR fp)
{
   FILE_POS     fp_pos;
   LOAD_PAT_PMT arg = {&(sys.module[sys.PMT_idx]),
                       0, 0x00, 0xFF};

   arg.PID = sys.get_pid(sys.psi_pes[sys.PAT_idx].common.mem,
                         sys.program_number);
   if (arg.PID < TS_PKT_NUL_ID) {
      FILE_GETPOS(fp, fp_pos);
      sys.PMT_changed(sys.psi_pes[sys.PMT_idx].common.mem);
      MPEG_TS_decode(fp, (fp_pos+CMD_SEARCH_SIZE), load_pat_pmt, &arg);
      FILE_SETPOS(fp, fp_pos);
   }
}

/* Decodes the MPEG transport stream in the given file
      INPUT: fp      - Pointer to an open file that contains MPEG TS 
                         data.
             end_pos - Point at which the decoding will stop. Decoding
                         will stop at an earlier point, if an EOF is
                         encountered.
             func    - Function that will deal with the decoded 
                        TS_PKTs.
             arg     - Pointer to any private data for func. */
void MPEG_TS_decode(FILE_PTR            fp, 
                    FILE_POS            end_pos, 
                    MPEG_TS_decode_func func,
                    void               *arg)
{
   FILE_POS        cur_pos;
   FILE_BYTE_COUNT bytes_read;
   TS_PKT          ts;
   u8              pkt[TS_PKT_SIZE];
   u32             buf_size, pkt_size;
   const u8       *ptr, *ptr2, *buf;
   
   ptr2     = NULL;
   pkt_size = 0;
   do {
      FILE_READ(fp, sys.buf, BUFFER_SIZE, bytes_read);
      if (bytes_read < 1)
         break;

      buf_size = (u32)bytes_read;
      ptr      = sys.buf;
      if ( (pkt_size)&&((pkt_size+buf_size) >= TS_PKT_SIZE) ) {
         /* complete the last packet */
         memcpy((pkt+pkt_size), ptr, (size_t)(TS_PKT_SIZE-pkt_size));
         ptr2     = ptr + (TS_PKT_SIZE-pkt_size);
         pkt_size = buf_size - (TS_PKT_SIZE-pkt_size);
         ptr      = pkt;         /* process this packet now */
         buf_size = TS_PKT_SIZE;
      }
      
MPEG_TS_decode_PROCESS_PACKETS:
      buf = ptr;
      while (buf_size >= TS_PKT_SIZE) {
         if (*buf != TS_PKT_SYNC_BYTE) {
            buf = TS_PKT_find(buf, buf_size);
            if (!buf) {
               buf_size = 0;
               break;   /* this and all following packets
                           in the buffer are bad */
            }
            goto MPEG_TS_decode_PACKET_DONE;
         }

         buf = TS_PKT_read((buf+1), &ts);
         if (!buf) {
            buf = ptr + TS_PKT_SIZE;
            goto MPEG_TS_decode_PACKET_DONE;   /* this packet is bad, so skip it */
         }
         else if ( (ts.PID < TS_PKT_NUL_ID)&&
                   (!func(&ts, arg)) ) {
            bytes_read = 0;
            buf_size   = 0;
            ptr2       = NULL;
            break;
         }
/*********** STOP DOING SOMETHING WITH ts ************/

MPEG_TS_decode_PACKET_DONE:
         buf_size -= (u32)(buf - ptr);
         ptr       = buf;
      }

      if (ptr2) {
         /* process the rest of the packets in the buffer */
         ptr      = ptr2;
         ptr2     = NULL;
         buf_size = pkt_size;
         pkt_size = 0;
         goto MPEG_TS_decode_PROCESS_PACKETS;
      }
      else if ( (buf_size)&&(buf) ) {
         /* store first part of the incomplete packet */
         memcpy(pkt, buf, (size_t)buf_size);
         pkt_size = buf_size;
      }

      FILE_GETPOS(fp, cur_pos);
   } while ( (bytes_read == BUFFER_SIZE)&&(cur_pos < end_pos) );
}


static int get_PIDs(TS_PKT *ts, void *arg)
{
   u32       i;
   GET_PIDS *data = (GET_PIDS *)arg;
   
   for (i = 0; i < data->size; i++) {
      if (data->PID[i] == ts->PID)
         break;
   }

   if (i >= data->size) {
      /* found a new PID */
      data->PID[data->found] = ts->PID;
      data->found++;
      if (data->found >= data->size)
         return 0;   /* we're done! */
   }
   return 1;
}

static int stream_type(TS_PKT *ts, void *arg)
{
   STREAM_TYPE *data = (STREAM_TYPE *)arg;

   if (ts->PID == data->PID) {
      if ( (!data->is_PES)&&
           (ts->payload_unit_start_indicator)&&
           (ts->payload_len32 >= 3) ) {
         if ( (!ts->payload[0])&&(!ts->payload[1])&&
              (ts->payload[2] == 0x01) )
            data->is_PES = 1;
      }
         
      if (data->is_PES) {
         MPEG2_SYS_MODULE *pes = &(sys.module[sys.PES_idx]);
         
         pes->process_pkt(pes->common.mem, ts);
         if (data->mod_data.stream_type)
            return 0;   /* we're done! */
      }
   }
   return 1;
}

static int load_pat_pmt(TS_PKT *ts, void *arg)
{
   LOAD_PAT_PMT *data = (LOAD_PAT_PMT *)arg;

   if (ts->PID == data->PID) {   
      /* check for discontinuity in data stream */
      if (data->last_cont_cnt != 0xFF) {
         if (ts->continuity_counter != (data->cont_cnt&TS_PKT_CONT_CNT_MASK)) {
#ifdef MPEG2_SYS_CLK64
            if ( (ts->continuity_counter != data->last_cont_cnt)||
                 ( (af_ptr)&&(ADAPT_FIELD_DISCONTINUITY_FLAG(af_ptr)) ) ) {
#else
            if (ts->continuity_counter != data->last_cont_cnt) {
#endif
               /* counter discontinuity */
               data->cont_cnt = ts->continuity_counter;
               data->mod->common.discontinuity(data->mod->common.mem);
            }
            else
               return 1;   /* duplicate packet */
         }
      }
      else  /* this is the first packet with ts.PID == PID[j] */
         data->cont_cnt = ts->continuity_counter;

      data->last_cont_cnt = ts->continuity_counter;
      if (ts->payload)
         data->cont_cnt++;

      data->mod->process_pkt(data->mod->common.mem, ts);
      if (ts->PID == TS_PKT_PAT_ID) {
         if (sys.get_program(sys.psi_pes[sys.PAT_idx].common.mem, 0))
            return 0;   /* finished loading the PAT */
      }
      else {
         if (sys.PMT_changed(sys.psi_pes[sys.PMT_idx].common.mem))
            return 0;   /* finished loading the PMT */
      }
   }
   return 1;
}

static int create_program(TS_PKT *ts, void *arg)
{
   CREATE_PROGRAM *data = (CREATE_PROGRAM *)arg;

   if ( (ts->PID != data->vid_PID)&&
        (ts->PID != data->aud_PID)&&
        (ts->payload_unit_start_indicator)&&
        (ts->payload_len32 >= 4) ) {
      if ( (!ts->payload[0])&&(!ts->payload[1])&&
           (ts->payload[2] == 0x01) ) {
         if (data->vid_PID >= TS_PKT_NUL_ID) {
            if ( (0xDF < ts->payload[3])&&(ts->payload[3] < 0xF0) )
               data->vid_PID = ts->PID;
         }

         if (data->aud_PID >= TS_PKT_NUL_ID) {
            if ( ((0xBF < ts->payload[3])&&(ts->payload[3] < 0xE0))||
                 (ts->payload[3] == PES_ATSC_AC3_AUDIO) )
               data->aud_PID = ts->PID;
         }
      }
   }

   if ( (data->vid_PID < TS_PKT_NUL_ID)&&
        (data->aud_PID < TS_PKT_NUL_ID) )
      return 0;   /* we're done! */
   return 1;
}

static void handle_PAT_PKT(MPEG2_SYS_ARG *sys_arg)
{
   u16 PMT_PID = sys.get_pid(sys.psi_pes[sys.PAT_idx].common.mem,
                             sys.program_number);

   if ( ( (PMT_PID < TS_PKT_NUL_ID)||(sys.PID[sys.PMT_idx] < TS_PKT_NUL_ID) )&&
        (PMT_PID != sys.PID[sys.PMT_idx]) ) {
      sys.PID[sys.PMT_idx] = PMT_PID;
      if (PMT_PID >= TS_PKT_NUL_ID) {
         /* erase PMT data, and stop processing elementary streams */
         sys.module[sys.PMT_idx].common.done(sys.module[sys.PMT_idx].common.mem);
         if (sys.PMT_changed(sys.psi_pes[sys.PMT_idx].common.mem))
            handle_PMT_PKT(sys_arg);
      }
      last_cont_cnt[sys.PMT_idx] = 0xFF;
   }
}

static void handle_PMT_PKT(MPEG2_SYS_ARG *sys_arg)
{
   u32          i, j,
                wake = 0;
   PMT_SECTION *pmt  = sys.get_PMT(sys.psi_pes[sys.PMT_idx].common.mem);

#ifndef _NO_VIDEO_
   lock_MUTEX(sys.m, 1);
   if ( (!stop_working)&&(sys.data.command != CMD_STOP) ) {
/*
      printf("%s> waiting for other thread to stop...\n", 
             (sys_arg->si) ? "AUDIO" : "VIDEO");
*/
      /* stop the other thread */
      stop_working = 1;
      broadcast_CONDVAR(sys.cv);
      do {
         wait_CONDVAR(sys.cv, sys.m);
      } while (stop_working == 1);
/*      printf("%s> its stopped!\n", (sys_arg->si) ? "AUDIO" : "VIDEO");*/
   }
   unlock_MUTEX(sys.m);
#endif

   /* shutdown the elementary streams */
   for (i = sys_arg->si; i < sys_arg->fi; i++) {
      if ( (i != sys.PMT_idx)&&(i != sys.PAT_idx) ) {
         sys.PID[i]       = TS_PKT_NUL_ID;
         last_cont_cnt[i] = 0xFF;
         if (sys.module[i].common.done != NULL)
            sys.module[i].common.done(sys.module[i].common.mem);
      }
   }
   
   if ( (pmt)&&(sys.data.command != CMD_STOP) ) {
      lock_MUTEX(sys.m, 1);
      /* initialize the elementary streams */
      for (i = 0; i < (u32)pmt->elem_size; i++) {
         for (j = 0; j < MPEG2_SYS_PIDS; j++) {
            if ( (j != sys.PMT_idx)&&(j != sys.PAT_idx)&&
                 (sys.psi_pes[j].stream_init != NULL) ) {
               if ( (pmt->elem[i]->elementary_PID < TS_PKT_NUL_ID)&&
                    (sys.psi_pes[j].stream_init(sys.psi_pes[j].common.mem, pmt, i)) ) {
                  if ( (j < sys_arg->si)||(j >= sys_arg->fi) )
                     wake++;  /* must wake up the other thread */
                  sys.PID[j] = pmt->elem[i]->elementary_PID;
                  break;
               }
            }
         }
      }
#ifndef _NO_VIDEO_
      if ( (stop_working)&&(wake) ) {
/*
         printf("%s> waiting for other thread to start working...\n", 
                (sys_arg->si) ? "AUDIO" : "VIDEO");
*/
         broadcast_CONDVAR(sys.cv);    /* get the other thread to start working */
         do {
            wait_CONDVAR(sys.cv, sys.m);  /* wait for the other thread to start */
         } while (stop_working);
/*         printf("%s> let's go!\n", (sys_arg->si) ? "AUDIO" : "VIDEO");*/
      }
#endif
      unlock_MUTEX(sys.m);
   }
}

#ifdef MPEG2_SYS_CLK64
static void update_offset(ADAPT_FIELD *af_ptr)
{
   int i;

   if (sys.offset == 0)
      i = 8;   /* initial sync */
   else if (ADAPT_FIELD_DISCONTINUITY_FLAG(af_ptr))
      i = 4;   /* discontinuity in the PCR */
   else if (af_ptr->PCR < sys.last_PCR)
      i = 2;   /* PCR clock overflow */
   else if (af_ptr->PCR < MPEG2_prog_clk(sys.offset))
      i = 1;   /* MPEG2_TS_decode() is not working fast enough */
   else
      i = 0;   /* everything is cool */
/*
   if ( (MPEG2_SYS_PRINT_DATA)&&(i > 1) ) {
      printf("MPEG2_TS_decode: "); 
*
      printf("PCR=0x%011llX  SYS=0x%016llX  ", af_ptr->PCR,
                                               MPEG2_prog_clk(sys->offset));
*
      switch (i) {
      case 1:
         printf("TOO SLOW\n");
         break;
      case 2:
         printf("PCR OVERFLOW\n");
         break;
      case 4:
         printf("PCR DISCONTINUITY\n");
         break;
      case 8:
         printf("INITIAL PCR SYNC\n");
      };
   }
*/
   if (i) {
      /* re-sync the program clock*/ 
      if (i & (2|4))
         sys.offset += af_ptr->PCR - sys.last_PCR;
      else
         sys.offset  = MPEG2_get_offset(af_ptr->PCR);
         
      if (sys.offset == 0)
         sys.offset = 1;  /* offset value cannot be 0 after re-syncing */

      /* notify all modules of the new offset value */
      for (i = 0; i < MPEG2_SYS_PIDS; i++) {
         if ( (i != sys.PMT_idx)&&(i != sys.PAT_idx)&&
              (sys.module[i].common.set_offset != NULL) ) {
            sys.module[i].common.set_offset(sys.module[i].common.mem,
                                            sys.offset);
         }
      }
/*
      if (MPEG2_SYS_PRINT_DATA)
         printf("                                 newSYS=0x%016llX\n", MPEG2_prog_clk(sys->offset));
*/
   }
   sys.last_PCR = af_ptr->PCR;
}
#endif

