/***************************************************************************
                          SLICE.h  -  description
                             -------------------
   Stuff needed for handling MPEG2 slices.
                             -------------------
    begin                : Tue Oct 5 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __SLICE1_H__
#define __SLICE1_H__

#include "common/typedefs.h"
#include "MPEG2_SYS.h"
#include "PIC_HEADER.h"
#include "mpeg2_video/SEQ_HEADER.h"


#define FRAME_420_SIZE(wid, hgt)    ((3*(wid)*(hgt)) >> 1)

#define PIC_DISPLAYED      0x80000000
#define PIC_PREV_GROUP     0x40000000

#define MB_MOTION_FWD_BIT  0x80
#define MB_MOTION_BCK_BIT  0x40
#define MB_PATTERN_BIT     0x20
#define MB_INTRA_BIT       0x10
#define MB_QUANT_BIT       0x08
#define MB_MISSING_MV_BIT  0x01

#define MB_MOTION_FWD(ptr_to_SLICE1) \
   (ptr_to_SLICE1->mb_flags & 0x80)
#define MB_MOTION_BCK(ptr_to_SLICE1) \
   (ptr_to_SLICE1->mb_flags & 0x40)
#define MB_PATTERN(ptr_to_SLICE1) \
   (ptr_to_SLICE1->mb_flags & 0x20)
#define MB_INTRA(ptr_to_SLICE1) \
   (ptr_to_SLICE1->mb_flags & 0x10)
#define MB_QUANT(ptr_to_SLICE1) \
   (ptr_to_SLICE1->mb_flags & 0x08)


#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   s16 QFS[64],            /* DCT coefficients / block values */
       vector[2][2],       /* motion vectors           (s, t) */
       PMV[2][2],          /* motion vector predictors (s, t) */
       dct_pred[3],        /* intra DC coefficient predictors (cc) */
       signed_level;       /* stored DCT coefficient */

   u16 mb_wid, mb_hgt,     /* size of reconstructed picture in MB units */
       mb_row, mb_col,     /* location of MB in the bitstream */
       cur_row, cur_col;   /* location of MB in the frame buffer */
       
   SEQ_HEADER  *seq_hdr;
   PIC_HEADER1 *pic_hdr;

   CLK_VAL offset, cur_pts, 
           pes_pts, ref_ptsF, ref_ptsB;
       
   s32 move[2];
   u32 ref_tempF,
       ref_tempB,
       cur_temp,
       frame_wid,
       luma_size,
       frame_size;
   u8 *ref_buf,            /* buffer holding the 2 reference frames */
      *ref_frameF,         /* forward prediction reference frame */
      *ref_blkF,           /* current block in fwd ref frame */
      *ref_frameB,         /* backward prediction reference frame */
      *ref_blkB,           /* current block in bck ref frame */
      *pred_frame,         /* this is frame being created */
      *pred_blk,           /* current block in pred frame */
      *mv_blk[2];
      
   u8  bit_shift, data, q_scale_code,
       mb_flags, r, s, t, i, cc, dct_size,
       last_frame, half_pixel[2], buf[2];
} SLICE1;

/* Processes a slice in data_stream.
      INPUT: data_stream - Points to the data stream holding the slice.
                             This stream does not have to contain the
                             entire slice. Repeated calls to this
                             function can be used to read the slice
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: seq_hdr - Points to the SEQ_HEADER1 of the mpeg1 stream.
              pic_hdr - Points to the PIC_HEADER1 of the slice.
              cur_pos - Used by MPEG1_VIDEO_MODULE.
              code    - Used by MPEG1_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 SLICE1_read(const u8     *data_stream,
                      u32     data_len,
                      SLICE1 *slice,
                      u32    *cur_pos,
                      u16    *code);

#ifdef __cplusplus
};
#endif

#endif

