/***************************************************************************
                     DirectSound_sound.h  -  description
                             -------------------
   Sound module implemented using Direct Sound.
                             -------------------
    begin                : Wed Aug 3 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>
#include <string.h>
#include "win32_main/winmain.h"
#include <mmreg.h>
#include <dsound.h>

#include "common/Timer.h"

#define REQ_BUFSIZE     (32 * 1024)
#define MAX_CHANNELS     2

static LPDIRECTSOUND       lpDS      = NULL;
static LPDIRECTSOUNDBUFFER lpDS_buf1 = NULL,
                           lpDS_buf2 = NULL;
static DWORD               writePos  = 0,
                           start_play= 0,
                           playPos   = 0,
                           byte_rate = 0;   /* bytes/ms */

static u32  channels   = 0,
            rate       = 0,
            buf_size   = 0,
            buf_used   = 0,
            bytes_sent = 0,
            bytes_play = 0;
static s16 *buf        = NULL;

/* Prepares the sound device for the incoming audio stream.
      INPUT: num_ch      - Number of audio channels.
             sample_rate - Sample rate (Hz)
      RETURN: Success - Number of audio channels.
              Failure - 0 */
u32 SOUND_init(u32 num_ch, u32 sample_rate)
{
   WAVEFORMATEX format;
   DSBUFFERDESC desc;
   HRESULT      rc;

   if (!lpDS) {
      rc = DirectSoundCreate(NULL, &lpDS, NULL);
      if ( (rc != DS_OK)||(!num_ch) )
         goto SOUND_init_ERROR;

      rc = IDirectSound_SetCooperativeLevel(lpDS, win, DSSCL_PRIORITY);
      if (rc != DS_OK)
         goto SOUND_init_ERROR;
   }

   if (num_ch > 2)
      num_ch = MAX_CHANNELS;

   if (lpDS_buf2) {
      if (!num_ch)
         goto SOUND_init_ERROR;
      else if ( (num_ch == channels)&&(sample_rate == rate) )
         return num_ch;
      IDirectSoundBuffer_Stop(lpDS_buf2);
      IDirectSoundBuffer_Release(lpDS_buf2);
      IDirectSoundBuffer_Release(lpDS_buf1);
   }
   else if (!num_ch)
      return 0;

   format.wFormatTag      = WAVE_FORMAT_PCM;
   format.nChannels       = (WORD)num_ch;
   format.nSamplesPerSec  = (DWORD)sample_rate;
   format.nAvgBytesPerSec = (16/8)*format.nChannels*format.nSamplesPerSec;
   format.nBlockAlign     = (16/8)*format.nChannels;
   format.wBitsPerSample  = 16;
   format.cbSize          = 0;

   /* Create Primary Buffer */
   memset(&desc, 0, sizeof(DSBUFFERDESC));
   desc.dwSize        = sizeof(DSBUFFERDESC);
   desc.dwFlags       = DSBCAPS_PRIMARYBUFFER;
   desc.dwBufferBytes = 0;
   desc.lpwfxFormat   = NULL;
   lpDS_buf1          = NULL;

   rc = IDirectSound_CreateSoundBuffer(lpDS, &desc, &lpDS_buf1, NULL);
   if (rc != DS_OK)
      goto SOUND_init_ERROR;

   byte_rate = (DWORD)(sample_rate*sizeof(s16)*num_ch) / 1000;
   channels  = num_ch;
   rate      = sample_rate;
   IDirectSoundBuffer_SetFormat(lpDS_buf1, &format);

   /* Create Secondary Buffer */
   bytes_sent = bytes_play = 0;
   start_play = 1;

   desc.dwFlags       = DSBCAPS_STICKYFOCUS |
                        DSBCAPS_GLOBALFOCUS |
                        DSBCAPS_STATIC |
                        DSBCAPS_GETCURRENTPOSITION2;
   desc.dwBufferBytes = REQ_BUFSIZE;
   desc.lpwfxFormat   = &format;
   lpDS_buf2          = NULL;
   rc = IDirectSound_CreateSoundBuffer(lpDS, &desc, &lpDS_buf2, NULL);
   if (rc == DS_OK) {
      IDirectSoundBuffer_GetCurrentPosition(lpDS_buf2, &playPos, &writePos);
      return num_ch;
   }

SOUND_init_ERROR:
   SOUND_done();
   return 0;
}

/* Returns the buffer that will hold the audio samples. Must
     call SOUND_init() function before calling this one.
      INPUT: samples - Number of samples the buffer will hold.
      RETURN:  Success - Pointer to the buffer.
               Failure - NULL */
s16 *SOUND_buf(u32 samples)
{
   if (channels) {
      samples *= sizeof(s16) * channels;
      if (samples > buf_size) {
         /* create a bigger buffer */
         if (buf)
            free(buf);
         buf      = malloc((size_t)samples);
         buf_size = (buf) ? samples : 0;
      }

      if (buf) {
         buf_used = samples;
         memset(buf, 0, (size_t)samples);
      }
      else
         buf_used = 0;
      return buf;
   }
   return NULL;
}

/* Called when the sound is no longer needed. */
void SOUND_done()
{
   if (lpDS_buf2) {
      if (buf_size) {
         u32   bytes_left;
         DWORD status = 0;

         IDirectSoundBuffer_GetStatus(lpDS_buf2, &status);
         if (status & DSBSTATUS_LOOPING) {
            memset(buf, 0, (size_t)buf_used);
            SOUND_play(); SOUND_play();
            bytes_left = SOUND_samples_left() * sizeof(s16) * channels;
            if (bytes_left > buf_used) {
               bytes_left -= buf_used;
               bytes_left /= byte_rate;
               TIMER_sleep(bytes_left+1); /* wait for last samples to play */
            }
         }
      }
      IDirectSoundBuffer_Stop(lpDS_buf2);
      IDirectSoundBuffer_Release(lpDS_buf2);
      lpDS_buf2 = NULL;
   }
   if (lpDS_buf1) {
      IDirectSoundBuffer_Release(lpDS_buf1);
      lpDS_buf1 = NULL;
      channels  = rate = 0;
   }
   if (lpDS) {
      IDirectSound_Release(lpDS);
      lpDS = NULL;
   }
   if (buf) {
      free(buf);
      buf      = NULL;
      buf_size = 0;
   }
}

/* Called when the samples in the buffer should be played. */
void SOUND_play()
{
   void   *ptr1, *ptr2;
   DWORD   size1, size2, bytes;
   HRESULT rc;
   u8     *sample_buf = (u8 *)buf;
   DWORD   play_bytes = (DWORD)buf_used;

SOUND_play_START:
   if ( (!start_play)||((bytes_sent + (u32)play_bytes) >= REQ_BUFSIZE) ) {
      if (start_play) {
         start_play = 0;
         IDirectSoundBuffer_Play(lpDS_buf2, 0, 0, DSBPLAY_LOOPING);
         bytes_play = 0;
      }

      size1 = playPos;
      IDirectSoundBuffer_GetCurrentPosition(lpDS_buf2,
                                            &playPos, &size2);
      bytes = playPos;
      if (playPos < size1)
         bytes += REQ_BUFSIZE;
      bytes_play += (u32)(bytes - size1);
      if (bytes_play >= bytes_sent) {
         writePos += (DWORD)(bytes_play - bytes_sent);
         if (writePos >= REQ_BUFSIZE)
            writePos -= REQ_BUFSIZE;
         bytes_sent = bytes_play;
         bytes      = REQ_BUFSIZE;     /* entire buffer is empty */
      }
      else {
         size2 = (DWORD)(bytes_sent - bytes_play);
         if (size2 >= (REQ_BUFSIZE-1024)) {
            if (size2 > REQ_BUFSIZE)
               bytes_play = bytes_sent - REQ_BUFSIZE;
            TIMER_sleep((REQ_BUFSIZE>>2)/byte_rate);
            goto SOUND_play_START;     /* buffer is full */
         }
         else
            bytes = REQ_BUFSIZE - size2;  /* free bytes */
      }
      if (bytes > play_bytes)
         bytes = play_bytes;
   }
   else if ((REQ_BUFSIZE-bytes_sent) > (u32)play_bytes)
      bytes = play_bytes;
   else
      bytes = (DWORD)(REQ_BUFSIZE-bytes_sent);

SOUND_play_LOCK_BUFFER:
   rc = IDirectSoundBuffer_Lock(lpDS_buf2, writePos, bytes,
                                &ptr1, &size1, &ptr2, &size2, 0);
   if (rc != DS_OK) {
      if ( (rc == DSERR_BUFFERLOST)&&(start_play < 20) ) {
         IDirectSoundBuffer_Restore(lpDS_buf2);
         start_play++;
         goto SOUND_play_LOCK_BUFFER;
      }
      else {
         bytes_sent = bytes_play = 0;
         start_play = 1;
         writePos   = playPos = 0;
         IDirectSoundBuffer_SetCurrentPosition(lpDS_buf2, 0);
         goto SOUND_play_START;
      }
   }

   memcpy(ptr1, sample_buf, (size_t)size1);
   if (ptr2)
      memcpy(ptr2, (sample_buf + size1), (size_t)size2);

   IDirectSoundBuffer_Unlock(lpDS_buf2,
                             ptr1, size1, ptr2, size2);

   play_bytes -= bytes;
   bytes_sent += (u32)bytes;
   writePos   += bytes;
   if (writePos >= REQ_BUFSIZE)
      writePos -= REQ_BUFSIZE;
   if (play_bytes) {
      sample_buf += bytes;
      goto SOUND_play_START;
   }
}

/* Returns the number of samples left in the buffer. */
u32 SOUND_samples_left()
{
   DWORD size1, size2, bytes;

   size1 = playPos;
   IDirectSoundBuffer_GetCurrentPosition(lpDS_buf2,
                                         &playPos, &size2);
   bytes = playPos;
   if (playPos < size1)
      bytes += REQ_BUFSIZE;
   bytes_play += (u32)(bytes - size1);
   if (bytes_play >= bytes_sent) {
      writePos += (DWORD)(bytes_play - bytes_sent);
      if (writePos >= REQ_BUFSIZE)
         writePos -= REQ_BUFSIZE;
      bytes_sent = bytes_play;
      bytes      = 0;    /* entire buffer is empty */
   }
   else {
      bytes = (DWORD)(bytes_sent - bytes_play);
      if (bytes > REQ_BUFSIZE) {
         bytes      = REQ_BUFSIZE;
         bytes_play = bytes_sent - REQ_BUFSIZE;
      }
      else if (bytes <= 512)
         bytes  = 0;
      else
         bytes -= 512;
   }
   return ((u32)bytes / (sizeof(s16) * channels));
}

/* This function is used to explicitly stop the sound from
    being heard.  Continuously running systems (ie. DirectSound)
    need this to prevent users from hearing annoying looping
    sounds when there are no new samples being sent to the
    sound card (during a pause).
      INPUT: wait - Set when you want the sound to start after
                     this function is called again with wait=0.
                    Unset when you want the sound to start at
                     the next SOUND_play() call. */
void SOUND_stop(int wait)
{
   if (lpDS_buf2) {
      IDirectSoundBuffer_Stop(lpDS_buf2);

      if (!wait) {
         bytes_sent = bytes_play = 0;
         start_play = 1;
         IDirectSoundBuffer_GetCurrentPosition(lpDS_buf2, &playPos, &writePos);
      }
   }
}

#undef REQ_BUFSIZE
#undef MAX_CHANNELS
