/***************************************************************************
                         StatsDLG.c  -  description
                             -------------------
   Stuff needed for handling the decoding stats dialog.
                             -------------------
    begin                : Tue Feb 22 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#ifndef WIN32_LEAN_AND_MEAN
# define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>

#include "winmain.h"
#include "resource.h"
#include "StatsDLG.h"
#include "MPEG2_SYS.h"


static BOOL CALLBACK StatsDLG_Proc(HWND hw, UINT msg,
                                   WPARAM p1, LPARAM p2);

void StatsDLG()
{
   DialogBox(app, MAKEINTRESOURCE(IDD_STATS_DIALOG), 
             win, StatsDLG_Proc);
}


static void format_time(char *txt, int hrs, int min, double sec)
{
   if (hrs > 0) {
      sprintf(txt, " %dh%02dm%06.3lfsec",
              hrs, (min - (hrs*60)),
              (sec - (double)(min*60)));
   }
   else if (min > 0) {
      sprintf(txt, " %dm%06.3lfsec",
              min, (sec - (double)(min*60)));
   }
   else if (sec >= 1.0)
      sprintf(txt, " %.3lfsec", sec);
   else
      sprintf(txt, " %.3lfmsec", (1.0e3*sec));
}

static BOOL CALLBACK StatsDLG_Proc(HWND hw, UINT msg,
                                   WPARAM p1, LPARAM p2)
{
   switch (msg) {
   case WM_COMMAND:
      if ( (HIWORD(p1) == BN_CLICKED)&&(LOWORD(p1) == IDOK) )
         EndDialog(hw, 0);
      break;
   case WM_INITDIALOG:
   {
      char txt[64];

      /* display stream length */
      if (sys.data.v_pos > 0) {
         format_time(txt,  
                     (int)(sys.data.v_pos/(  3600*MPEG2_SYS_CLK_FREQ)),
                     (int)(sys.data.v_pos/(    60*MPEG2_SYS_CLK_FREQ)),
                   (double)sys.data.v_pos/(double)MPEG2_SYS_CLK_FREQ);
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_VID_TIME, 
                            WM_SETTEXT, 0, (LPARAM)txt);
      }
      else {
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_VID_TIME, 
                            WM_SETTEXT, 0, (LPARAM)" N/A");
      }
      if (sys.data.a_pos > 0) {
         format_time(txt,
                     (int)(sys.data.a_pos/(  3600*MPEG2_SYS_CLK_FREQ)),
                     (int)(sys.data.a_pos/(    60*MPEG2_SYS_CLK_FREQ)),
                   (double)sys.data.a_pos/(double)MPEG2_SYS_CLK_FREQ);
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_AUD_TIME, 
                            WM_SETTEXT, 0, (LPARAM)txt);
      }
      else {
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_AUD_TIME, 
                            WM_SETTEXT, 0, (LPARAM)" N/A");
      }

#if (MPEG2_SYS_SLOW_DOWN > 1)
      sys.decode_time /= MPEG2_SYS_SLOW_DOWN;
#endif
      if (sys.decode_time > 0.0) {
         double rate = (double)(8*sys.stream_bytes) / sys.decode_time;

         /* display decoding time */
         format_time(txt,
                     (int)(sys.decode_time / 3600),
                     (int)(sys.decode_time /   60),
                     sys.decode_time);
#if (MPEG2_SYS_SLOW_DOWN > 1)
         sprintf((txt + strlen(txt)), " (x%d)", MPEG2_SYS_SLOW_DOWN);
#endif
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_DEC_TIME, 
                            WM_SETTEXT, 0, (LPARAM)txt);

         /* display bit rate */
         if (rate > 1.0e9)
            sprintf(txt, " %.3lfGbps", (rate / 1.0e9));
         else if (rate > 1.0e6)
            sprintf(txt, " %.3lfMbps", (rate / 1.0e6));
         else if (rate > 1.0e3)
            sprintf(txt, " %.3lfKbps", (rate / 1.0e3));
         else
            sprintf(txt, " %.3lfbps", rate);
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_BIT_RATE, 
                            WM_SETTEXT, 0, (LPARAM)txt);

         if (sys.pkt_cnt) {
            /* display packet count */
            sprintf(txt, " %u", sys.pkt_cnt);
            SendDlgItemMessage(hw, IDC_STATS_DIALOG_PACKETS, 
                               WM_SETTEXT, 0, (LPARAM)txt);

            /* display packet processing time */
            rate = sys.decode_time / (double)sys.pkt_cnt;
            if (rate > 1.0)
               sprintf(txt, " %.3lfsec/pkt", rate);
            else if (rate > 1.0e-3)
               sprintf(txt, " %.3lfmsec/pkt", (rate * 1.0e3));
            else
               sprintf(txt, " %.3lfnsec/pkt", (rate * 1.0e6));
            SendDlgItemMessage(hw, IDC_STATS_DIALOG_PACKET_TIME, 
                               WM_SETTEXT, 0, (LPARAM)txt);
         }
         else {
            /* packet info is not available */
            SendDlgItemMessage(hw, IDC_STATS_DIALOG_PACKETS, 
                               WM_SETTEXT, 0, (LPARAM)" N/A");
            SendDlgItemMessage(hw, IDC_STATS_DIALOG_PACKET_TIME, 
                               WM_SETTEXT, 0, (LPARAM)" N/A");
         }
      }
      else {
         /* no stats were recorded */
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_DEC_TIME, 
                            WM_SETTEXT, 0, (LPARAM)" N/A");
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_BIT_RATE, 
                            WM_SETTEXT, 0, (LPARAM)" N/A");
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_PACKETS, 
                            WM_SETTEXT, 0, (LPARAM)" N/A");
         SendDlgItemMessage(hw, IDC_STATS_DIALOG_PACKET_TIME, 
                            WM_SETTEXT, 0, (LPARAM)" N/A");
      }
      return TRUE;
   }
   }
   return FALSE;
}

